"""
A 2DPackLib instance.

This module provides an instance of the two-dimensional bin packing problem
as defined in 2DPackLib [1, 2].

All instances of :class:`~moptipyapps.binpacking2d.instance.Instance`
are two-dimensional numpy `ndarrays` with additional attributes.
Each instance has a :attr:`~Instance.name`. Instances also specify a
:attr:`~Instance.bin_width` and :attr:`~Instance.bin_height`.
They define the number :attr:`~Instance.n_different_items` of items with
*different* IDs. Notice that in the 2DPackLib dataset, a benchmark instance
may contain the same item multiple times. Obviously, all items of the same
ID have the exact same width and height, meaning that we only need to store
them once and remember how often they occur. (Notice that the opposite is not
true, i.e., not all items with the same width and height do have the same ID.)
Anyway, the total number :attr:`~Instance.n_items` of items, i.e., the sum of
all the repetitions of all items, is also stored.

The matrix data of the instance class is laid out as follows: There is one row
for each item. The row contains the width of the item, the height of the item,
and the number of times the item will occur. The row at index `i` stands for
the item with ID `i+1`.

Instances can be loaded directly from a 2DPackLib file via
:meth:`Instance.from_2dpacklib`. They can also be loaded from a compact string
representation (via :meth:`Instance.from_compact_str`) and can also be
converted to such a compact representation
(via :meth:`Instance.to_compact_str`). This library ships with a set of
pre-defined instances as resource which can be obtained via
:meth:`Instance.from_resource` and listed via
:meth:`Instance.list_resources`.

We provide the instances of the sets `A` [3], `BENG` [4], and `CLASS` [5].

Initial work on this code has been contributed by Mr. Rui ZHAO (赵睿),
<zr1329142665@163.com> a Master's student at the Institute of Applied
Optimization (应用优化研究所, http://iao.hfuu.edu.cn) of the School of
Artificial Intelligence and Big Data (人工智能与大数据学院) at Hefei University
(合肥学院) in Hefei, Anhui, China (中国安徽省合肥市) under the supervision of
Prof. Dr. Thomas Weise (汤卫思教授).

1. Manuel Iori, Vinícius Loti de Lima, Silvano Martello, and Michele Monaci.
   *2DPackLib*.
   https://site.unibo.it/operations-research/en/research/2dpacklib
2. Manuel Iori, Vinícius Loti de Lima, Silvano Martello, and Michele
   Monaci. 2DPackLib: A Two-Dimensional Cutting and Packing Library.
   *Optimization Letters* 16(2):471-480. March 2022.
   https://doi.org/10.1007/s11590-021-01808-y
3. Rita Macedo, Cláudio Alves, and José M. Valério de Carvalho. Arc-Flow
   Model for the Two-Dimensional Guillotine Cutting Stock Problem.
   *Computers & Operations Research* 37(6):991-1001. June 2010.
   https://doi.org/10.1016/j.cor.2009.08.005.
4. Bengt-Erik Bengtsson. Packing Rectangular Pieces - A Heuristic Approach.
   *The Computer Journal* 25(3):353-357, August 1982.
   https://doi.org/10.1093/comjnl/25.3.353
5. J.O. Berkey and P.Y. Wang. Two Dimensional Finite Bin Packing Algorithms.
   *Journal of the Operational Research Society* 38(5):423-429. May 1987.
   https://doi.org/10.1057/jors.1987.70

>>> ins = Instance("a", 100, 50, [[10, 5, 1], [3, 3, 1], [5, 5, 1]])
>>> ins.name
'a'
>>> ins.bin_width
100
>>> ins.bin_height
50
>>> ins.dtype
dtype('int8')
>>> ins.n_different_items
3
>>> ins.to_compact_str()
'a;3;100;50;10,5;3,3;5,5'
>>> ins = Instance("b", 100, 50, np.array([[10, 5, 1], [3, 3, 1], [3, 3, 1]]))
>>> ins.name
'b'
>>> ins.bin_width
100
>>> ins.bin_height
50
>>> ins.dtype
dtype('int8')
>>> ins.to_compact_str()
'b;3;100;50;10,5;3,3;3,3'
>>> ins = Instance.from_resource("cl02_020_06")
>>> ins.dtype
dtype('int8')
>>> ins.n_different_items
20
>>> ins.n_items
20
>>> ins = Instance.from_resource("a25")
>>> ins.dtype
dtype('int16')
>>> ins.n_different_items
75
>>> ins.n_items
156
"""

from importlib import resources  # nosem
from os.path import basename
from typing import Final, cast

import moptipy.utils.nputils as npu
import numpy as np
from moptipy.api.component import Component
from moptipy.utils.logger import CSV_SEPARATOR, KeyValueLogSection
from moptipy.utils.nputils import int_range_to_dtype
from moptipy.utils.path import Path
from moptipy.utils.strings import sanitize_name
from moptipy.utils.types import check_int_range, check_to_int_range, type_error

#: the instances resource name
INSTANCES_RESOURCE: Final[str] = "2dpacklib.txt"

#: the total number of items
N_ITEMS: Final[str] = "nItems"
#: the number of different items.
N_DIFFERENT_ITEMS: Final[str] = "nDifferentItems"
#: the bin width
BIN_WIDTH: Final[str] = "binWidth"
#: the bin height
BIN_HEIGHT: Final[str] = "binHeight"
#: The internal item separator
INTERNAL_SEP: Final[str] = "," if CSV_SEPARATOR == ";" else ";"

#: the index of the width element in an item of an instance
IDX_WIDTH: Final[int] = 0
#: the index of the height element in an item of an instance
IDX_HEIGHT: Final[int] = 1
#: the index of the repetitions element in an item of an instance
IDX_REPETITION: Final[int] = 2

#: the list of instance names of the 2DPackLib bin packing set
_INSTANCES: Final[tuple[str, ...]] = (
    "a01", "a02", "a03", "a04", "a05", "a06", "a07", "a08", "a09", "a10",
    "a11", "a12", "a13", "a14", "a15", "a16", "a17", "a18", "a19", "a20",
    "a21", "a22", "a23", "a24", "a25", "a26", "a27", "a28", "a29", "a30",
    "a31", "a32", "a33", "a34", "a35", "a36", "a37", "a38", "a39", "a40",
    "a41", "a42", "a43", "beng01", "beng02", "beng03", "beng04", "beng05",
    "beng06", "beng07", "beng08", "beng09", "beng10", "cl01_020_01",
    "cl01_020_02", "cl01_020_03", "cl01_020_04", "cl01_020_05", "cl01_020_06",
    "cl01_020_07", "cl01_020_08", "cl01_020_09", "cl01_020_10", "cl01_040_01",
    "cl01_040_02", "cl01_040_03", "cl01_040_04", "cl01_040_05", "cl01_040_06",
    "cl01_040_07", "cl01_040_08", "cl01_040_09", "cl01_040_10", "cl01_060_01",
    "cl01_060_02", "cl01_060_03", "cl01_060_04", "cl01_060_05", "cl01_060_06",
    "cl01_060_07", "cl01_060_08", "cl01_060_09", "cl01_060_10", "cl01_080_01",
    "cl01_080_02", "cl01_080_03", "cl01_080_04", "cl01_080_05", "cl01_080_06",
    "cl01_080_07", "cl01_080_08", "cl01_080_09", "cl01_080_10", "cl01_100_01",
    "cl01_100_02", "cl01_100_03", "cl01_100_04", "cl01_100_05", "cl01_100_06",
    "cl01_100_07", "cl01_100_08", "cl01_100_09", "cl01_100_10", "cl02_020_01",
    "cl02_020_02", "cl02_020_03", "cl02_020_04", "cl02_020_05", "cl02_020_06",
    "cl02_020_07", "cl02_020_08", "cl02_020_09", "cl02_020_10", "cl02_040_01",
    "cl02_040_02", "cl02_040_03", "cl02_040_04", "cl02_040_05", "cl02_040_06",
    "cl02_040_07", "cl02_040_08", "cl02_040_09", "cl02_040_10", "cl02_060_01",
    "cl02_060_02", "cl02_060_03", "cl02_060_04", "cl02_060_05", "cl02_060_06",
    "cl02_060_07", "cl02_060_08", "cl02_060_09", "cl02_060_10", "cl02_080_01",
    "cl02_080_02", "cl02_080_03", "cl02_080_04", "cl02_080_05", "cl02_080_06",
    "cl02_080_07", "cl02_080_08", "cl02_080_09", "cl02_080_10", "cl02_100_01",
    "cl02_100_02", "cl02_100_03", "cl02_100_04", "cl02_100_05", "cl02_100_06",
    "cl02_100_07", "cl02_100_08", "cl02_100_09", "cl02_100_10", "cl03_020_01",
    "cl03_020_02", "cl03_020_03", "cl03_020_04", "cl03_020_05", "cl03_020_06",
    "cl03_020_07", "cl03_020_08", "cl03_020_09", "cl03_020_10", "cl03_040_01",
    "cl03_040_02", "cl03_040_03", "cl03_040_04", "cl03_040_05", "cl03_040_06",
    "cl03_040_07", "cl03_040_08", "cl03_040_09", "cl03_040_10", "cl03_060_01",
    "cl03_060_02", "cl03_060_03", "cl03_060_04", "cl03_060_05", "cl03_060_06",
    "cl03_060_07", "cl03_060_08", "cl03_060_09", "cl03_060_10", "cl03_080_01",
    "cl03_080_02", "cl03_080_03", "cl03_080_04", "cl03_080_05", "cl03_080_06",
    "cl03_080_07", "cl03_080_08", "cl03_080_09", "cl03_080_10", "cl03_100_01",
    "cl03_100_02", "cl03_100_03", "cl03_100_04", "cl03_100_05", "cl03_100_06",
    "cl03_100_07", "cl03_100_08", "cl03_100_09", "cl03_100_10", "cl04_020_01",
    "cl04_020_02", "cl04_020_03", "cl04_020_04", "cl04_020_05", "cl04_020_06",
    "cl04_020_07", "cl04_020_08", "cl04_020_09", "cl04_020_10", "cl04_040_01",
    "cl04_040_02", "cl04_040_03", "cl04_040_04", "cl04_040_05", "cl04_040_06",
    "cl04_040_07", "cl04_040_08", "cl04_040_09", "cl04_040_10", "cl04_060_01",
    "cl04_060_02", "cl04_060_03", "cl04_060_04", "cl04_060_05", "cl04_060_06",
    "cl04_060_07", "cl04_060_08", "cl04_060_09", "cl04_060_10", "cl04_080_01",
    "cl04_080_02", "cl04_080_03", "cl04_080_04", "cl04_080_05", "cl04_080_06",
    "cl04_080_07", "cl04_080_08", "cl04_080_09", "cl04_080_10", "cl04_100_01",
    "cl04_100_02", "cl04_100_03", "cl04_100_04", "cl04_100_05", "cl04_100_06",
    "cl04_100_07", "cl04_100_08", "cl04_100_09", "cl04_100_10", "cl05_020_01",
    "cl05_020_02", "cl05_020_03", "cl05_020_04", "cl05_020_05", "cl05_020_06",
    "cl05_020_07", "cl05_020_08", "cl05_020_09", "cl05_020_10", "cl05_040_01",
    "cl05_040_02", "cl05_040_03", "cl05_040_04", "cl05_040_05", "cl05_040_06",
    "cl05_040_07", "cl05_040_08", "cl05_040_09", "cl05_040_10", "cl05_060_01",
    "cl05_060_02", "cl05_060_03", "cl05_060_04", "cl05_060_05", "cl05_060_06",
    "cl05_060_07", "cl05_060_08", "cl05_060_09", "cl05_060_10", "cl05_080_01",
    "cl05_080_02", "cl05_080_03", "cl05_080_04", "cl05_080_05", "cl05_080_06",
    "cl05_080_07", "cl05_080_08", "cl05_080_09", "cl05_080_10", "cl05_100_01",
    "cl05_100_02", "cl05_100_03", "cl05_100_04", "cl05_100_05", "cl05_100_06",
    "cl05_100_07", "cl05_100_08", "cl05_100_09", "cl05_100_10", "cl06_020_01",
    "cl06_020_02", "cl06_020_03", "cl06_020_04", "cl06_020_05", "cl06_020_06",
    "cl06_020_07", "cl06_020_08", "cl06_020_09", "cl06_020_10", "cl06_040_01",
    "cl06_040_02", "cl06_040_03", "cl06_040_04", "cl06_040_05", "cl06_040_06",
    "cl06_040_07", "cl06_040_08", "cl06_040_09", "cl06_040_10", "cl06_060_01",
    "cl06_060_02", "cl06_060_03", "cl06_060_04", "cl06_060_05", "cl06_060_06",
    "cl06_060_07", "cl06_060_08", "cl06_060_09", "cl06_060_10", "cl06_080_01",
    "cl06_080_02", "cl06_080_03", "cl06_080_04", "cl06_080_05", "cl06_080_06",
    "cl06_080_07", "cl06_080_08", "cl06_080_09", "cl06_080_10", "cl06_100_01",
    "cl06_100_02", "cl06_100_03", "cl06_100_04", "cl06_100_05", "cl06_100_06",
    "cl06_100_07", "cl06_100_08", "cl06_100_09", "cl06_100_10", "cl07_020_01",
    "cl07_020_02", "cl07_020_03", "cl07_020_04", "cl07_020_05", "cl07_020_06",
    "cl07_020_07", "cl07_020_08", "cl07_020_09", "cl07_020_10", "cl07_040_01",
    "cl07_040_02", "cl07_040_03", "cl07_040_04", "cl07_040_05", "cl07_040_06",
    "cl07_040_07", "cl07_040_08", "cl07_040_09", "cl07_040_10", "cl07_060_01",
    "cl07_060_02", "cl07_060_03", "cl07_060_04", "cl07_060_05", "cl07_060_06",
    "cl07_060_07", "cl07_060_08", "cl07_060_09", "cl07_060_10", "cl07_080_01",
    "cl07_080_02", "cl07_080_03", "cl07_080_04", "cl07_080_05", "cl07_080_06",
    "cl07_080_07", "cl07_080_08", "cl07_080_09", "cl07_080_10", "cl07_100_01",
    "cl07_100_02", "cl07_100_03", "cl07_100_04", "cl07_100_05", "cl07_100_06",
    "cl07_100_07", "cl07_100_08", "cl07_100_09", "cl07_100_10", "cl08_020_01",
    "cl08_020_02", "cl08_020_03", "cl08_020_04", "cl08_020_05", "cl08_020_06",
    "cl08_020_07", "cl08_020_08", "cl08_020_09", "cl08_020_10", "cl08_040_01",
    "cl08_040_02", "cl08_040_03", "cl08_040_04", "cl08_040_05", "cl08_040_06",
    "cl08_040_07", "cl08_040_08", "cl08_040_09", "cl08_040_10", "cl08_060_01",
    "cl08_060_02", "cl08_060_03", "cl08_060_04", "cl08_060_05", "cl08_060_06",
    "cl08_060_07", "cl08_060_08", "cl08_060_09", "cl08_060_10", "cl08_080_01",
    "cl08_080_02", "cl08_080_03", "cl08_080_04", "cl08_080_05", "cl08_080_06",
    "cl08_080_07", "cl08_080_08", "cl08_080_09", "cl08_080_10", "cl08_100_01",
    "cl08_100_02", "cl08_100_03", "cl08_100_04", "cl08_100_05", "cl08_100_06",
    "cl08_100_07", "cl08_100_08", "cl08_100_09", "cl08_100_10", "cl09_020_01",
    "cl09_020_02", "cl09_020_03", "cl09_020_04", "cl09_020_05", "cl09_020_06",
    "cl09_020_07", "cl09_020_08", "cl09_020_09", "cl09_020_10", "cl09_040_01",
    "cl09_040_02", "cl09_040_03", "cl09_040_04", "cl09_040_05", "cl09_040_06",
    "cl09_040_07", "cl09_040_08", "cl09_040_09", "cl09_040_10", "cl09_060_01",
    "cl09_060_02", "cl09_060_03", "cl09_060_04", "cl09_060_05", "cl09_060_06",
    "cl09_060_07", "cl09_060_08", "cl09_060_09", "cl09_060_10", "cl09_080_01",
    "cl09_080_02", "cl09_080_03", "cl09_080_04", "cl09_080_05", "cl09_080_06",
    "cl09_080_07", "cl09_080_08", "cl09_080_09", "cl09_080_10", "cl09_100_01",
    "cl09_100_02", "cl09_100_03", "cl09_100_04", "cl09_100_05", "cl09_100_06",
    "cl09_100_07", "cl09_100_08", "cl09_100_09", "cl09_100_10", "cl10_020_01",
    "cl10_020_02", "cl10_020_03", "cl10_020_04", "cl10_020_05", "cl10_020_06",
    "cl10_020_07", "cl10_020_08", "cl10_020_09", "cl10_020_10", "cl10_040_01",
    "cl10_040_02", "cl10_040_03", "cl10_040_04", "cl10_040_05", "cl10_040_06",
    "cl10_040_07", "cl10_040_08", "cl10_040_09", "cl10_040_10", "cl10_060_01",
    "cl10_060_02", "cl10_060_03", "cl10_060_04", "cl10_060_05", "cl10_060_06",
    "cl10_060_07", "cl10_060_08", "cl10_060_09", "cl10_060_10", "cl10_080_01",
    "cl10_080_02", "cl10_080_03", "cl10_080_04", "cl10_080_05", "cl10_080_06",
    "cl10_080_07", "cl10_080_08", "cl10_080_09", "cl10_080_10", "cl10_100_01",
    "cl10_100_02", "cl10_100_03", "cl10_100_04", "cl10_100_05", "cl10_100_06",
    "cl10_100_07", "cl10_100_08", "cl10_100_09", "cl10_100_10")


class Instance(Component, np.ndarray):
    """
    An instance of the 2D Bin Packing Problem.

    Each row of the matrix contains three values: 1. the item's width, 2. the
    item's height, 3. how often the item occurs.
    """

    #: the name of the instance
    name: str
    #: the total number of items (including repetitions)
    n_items: int
    #: the number of different items
    n_different_items: int
    #: the bin width
    bin_width: int
    #: the bin height
    bin_height: int
    #: the minimum number of bins that this instance requires
    lower_bound_bins: int

    def __new__(cls, name: str,
                bin_width: int, bin_height: int,
                matrix: np.ndarray | list[list[int]]) -> "Instance":
        """
        Create an instance of the 2D bin packing problem.

        :param cls: the class
        :param name: the name of the instance
        :param bin_width: the bin width
        :param bin_height: the bin height
        :param matrix: the matrix with the data (will be copied)
        """
        use_name: Final[str] = sanitize_name(name)
        if name != use_name:
            raise ValueError(f"Name {name!r} is not a valid name.")

        check_int_range(bin_width, "bin_width", 1, 1_000_000_000_000)
        check_int_range(bin_height, "bin_height", 1, 1_000_000_000_000)
        max_dim: Final[int] = max(bin_width, bin_height)
        min_dim: Final[int] = min(bin_width, bin_height)

        n_different_items: Final[int] = check_int_range(
            len(matrix), "n_different_items", 1, 100_000_000)
        use_shape: Final[tuple[int, int]] = (n_different_items, 3)

        if isinstance(matrix, np.ndarray):
            if not npu.is_np_int(matrix.dtype):
                raise ValueError(
                    "Matrix must have an integer type, but is of type "
                    f"{str(matrix.dtype)!r} in instance {name!r}.")
            if matrix.shape != use_shape:
                raise ValueError(
                    f"Invalid shape {str(matrix.shape)!r} of matrix: "
                    "must have three columns and two dimensions, must be "
                    f"equal to {use_shape} in instance {name!r}.")
        elif not isinstance(matrix, list):
            raise type_error(matrix, "matrix", np.ndarray)

        n_items: int = 0
        max_size: int = -1
        item_area: int = 0
        for i in range(n_different_items):
            row = matrix[i]
            if not isinstance(row, list | np.ndarray):
                raise type_error(
                    row, f"{row} at index {i} in {use_name!r}",
                    (list, np.ndarray))
            if len(row) != 3:
                raise ValueError(
                    f"invalid row {row} at index {i} in {use_name!r}.")
            width, height, repetitions = row
            width = check_int_range(int(width), "width", 1, max_dim)
            height = check_int_range(int(height), "height", 1, max_dim)
            repetitions = check_int_range(int(repetitions), "repetitions",
                                          1, 100_000_000)
            item_area += (width * height * repetitions)
            max_size = max(width, height)
            if (width > min_dim) and (height > min_dim):
                raise ValueError(
                    f"object with width={width} and height={height} does "
                    f"not fit into bin with width={width} and "
                    f"height={height}.")
            n_items += repetitions

        obj: Final[Instance] = super().__new__(
            cls, use_shape, int_range_to_dtype(
                min_value=0, max_value=max(max_dim + max_size, n_items),
                force_signed=True))
        for i in range(n_different_items):
            obj[i, :] = matrix[i]

        #: the name of the instance
        obj.name = use_name
        #: the number of different items
        obj.n_different_items = n_different_items
        #: the total number of items, i.e., the number of different items
        #: multiplied with their repetition counts
        obj.n_items = n_items
        #: the height of the bins
        obj.bin_height = bin_height
        #: the width of the bins
        obj.bin_width = bin_width

# We need as least as many bins such that their area is big enough
# for the total area of the items.
        bin_area: int = bin_height * bin_width
        min_size: int = item_area // bin_area
        if (min_size * bin_area) < item_area:
            min_size += 1
        #: the lower bound for the number of bins needed
        obj.lower_bound_bins = check_int_range(
            min_size, "lower_bound_bins", 1, 1_000_000_000_000)
        return obj

    def __str__(self):
        """
        Get the name of this instance.

        :return: the name of this instance
        """
        return self.name

    def get_standard_item_sequence(self) -> list[int]:
        """
        Get the standardized item sequence.

        :return: the standardized item sequence

        >>> ins = Instance("a", 100, 50, [[10, 5, 1], [3, 3, 1], [5, 5, 1]])
        >>> ins.get_standard_item_sequence()
        [1, 2, 3]
        >>> ins = Instance("a", 100, 50, [[10, 5, 1], [3, 3, 1], [5, 5, 2]])
        >>> ins.get_standard_item_sequence()
        [1, 2, 3, 3]
        >>> ins = Instance("a", 100, 50, [[10, 5, 2], [3, 3, 3], [5, 5, 4]])
        >>> ins.get_standard_item_sequence()
        [1, 1, 2, 2, 2, 3, 3, 3, 3]
        """
        base_string: Final[list[int]] = []
        for i in range(1, self.n_different_items + 1):
            for _ in range(self[i - 1, IDX_REPETITION]):
                base_string.append(int(i))
        return base_string

    @staticmethod
    def from_2dpacklib(file: str) -> "Instance":
        """
        Load a problem instance from the 2dpacklib format.

        :param file: the file path
        :return: the instance
        """
        path: Final[Path] = Path.file(file)
        name: str = basename(path).lower()
        if name.endswith(".ins2d"):
            name = sanitize_name(name[:-6])

        lines: Final[list[str]] = path.read_all_list()
        n_different_items: Final[int] = check_to_int_range(
            lines[0], "n_different_items", 1, 100_000_000)

        wh: Final[str] = lines[1]
        spaceidx: Final[int] = wh.index(" ")
        if spaceidx <= 0:
            raise ValueError("Did not find space in second line.")
        bin_width: Final[int] = check_to_int_range(
            wh[:spaceidx], "bin_width", 1, 1_000_000_000_000)
        bin_height: Final[int] = check_to_int_range(
            wh[spaceidx + 1:], "bin_height", 1, 1_000_000_000_000)
        del lines[0:2]

        max_dim: Final[int] = max(bin_width, bin_height)
        data: Final[list[list[int]]] = []
        old_item_id: int = 0
        for line in lines:
            text: list[str] = line.split(" ")
            itemid: int = check_to_int_range(
                text[0], "item-id", 1, n_different_items)
            if itemid != (old_item_id + 1):
                raise ValueError(
                    f"non-sequential item id {itemid} after {old_item_id}!")
            old_item_id = itemid
            width: int = check_to_int_range(text[1], "width", 1, max_dim)
            height: int = check_to_int_range(text[2], "height", 1, max_dim)
            count: int = 1 if len(text) <= 3 else \
                check_to_int_range(text[3], "count", 1, 1_000_000)
            data.append([width, height, count])
        data.sort()
        return Instance(name, bin_width, bin_height, data)

    def to_compact_str(self) -> str:
        """
        Convert the instance to a compact string.

        The format of the string is a single line of semi-colon separated
        values. The values are: `name;n_items;bin_width;bin_height`,
        followed by the sequence of items, each in the form of
        `;width,heigh[,times]`, where `times` is optional and only added
        if the item occurs more than once.

        :return: a single line string with all the instance data

        >>> ins = Instance("x", 500, 50, [[3, 5, 1], [2, 5, 2]])
        >>> ins.to_compact_str()
        'x;2;500;50;3,5;2,5,2'
        >>> ins.n_different_items
        2
        """
        lst: Final[list[str]] = [self.name, str(self.n_different_items),
                                 str(self.bin_width), str(self.bin_height)]
        for i in range(self.n_different_items):
            width: int = self[i, IDX_WIDTH]
            height: int = self[i, IDX_HEIGHT]
            repetitions: int = self[i, IDX_REPETITION]
            lst.append(
                f"{width}{INTERNAL_SEP}{height}" if repetitions == 1 else
                f"{width}{INTERNAL_SEP}{height}{INTERNAL_SEP}{repetitions}")
        return CSV_SEPARATOR.join(lst)

    @staticmethod
    def from_compact_str(data: str) -> "Instance":
        """
        Transform a compact string back to an instance.

        :param data: the string data
        :return: the instance

        >>> ins = Instance("x", 500, 50, [[3, 5, 1], [2, 5, 2]])
        >>> Instance.from_compact_str(ins.to_compact_str()).to_compact_str()
        'x;2;500;50;3,5;2,5,2'
        """
        if not isinstance(data, str):
            raise type_error(data, "data", str)
        text: Final[list[str]] = data.split(CSV_SEPARATOR)
        name: Final[str] = text[0]
        n_different_items: Final[int] = check_to_int_range(
            text[1], "n_different_items", 1, 100_000_000)
        bin_width: Final[int] = check_to_int_range(
            text[2], "bin_width", 1, 1_000_000_000_000)
        bin_height: Final[int] = check_to_int_range(
            text[3], "bin_height", 1, 1_000_000_000_000)
        max_dim: Final[int] = max(bin_width, bin_height)
        items: list[list[int]] = []
        for i in range(4, n_different_items + 4):
            s: list[str] = text[i].split(INTERNAL_SEP)
            row: list[int] = [
                check_to_int_range(s[IDX_WIDTH], "width", 1, max_dim),
                check_to_int_range(s[IDX_HEIGHT], "height", 1, max_dim),
                1 if len(s) <= IDX_REPETITION else
                check_to_int_range(
                    s[IDX_REPETITION], "times", 1, 100_000_000)]
            items.append(row)
        return Instance(name, bin_width, bin_height, items)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters describing this bin packing instance to the logger.

        :param logger: the logger for the parameters

        >>> from moptipy.utils.logger import InMemoryLogger
        >>> with InMemoryLogger() as l:
        ...     with l.key_values("I") as kv:
        ...         Instance.from_resource("beng05").log_parameters_to(kv)
        ...     print(repr('@'.join(l.get_log())))
        'BEGIN_I@name: beng05@class: moptipyapps.binpacking2d\
.instance.Instance@nItems: 100@nDifferentItems: 100@binWidth: 25\
@binHeight: 10@dtype: b@END_I'
        """
        super().log_parameters_to(logger)
        logger.key_value(N_ITEMS, self.n_items)
        logger.key_value(N_DIFFERENT_ITEMS, self.n_different_items)
        logger.key_value(BIN_WIDTH, self.bin_width)
        logger.key_value(BIN_HEIGHT, self.bin_height)
        logger.key_value(npu.KEY_NUMPY_TYPE, self.dtype.char)

    @staticmethod
    def list_resources() -> tuple[str, ...]:
        """
        Get a tuple of all the instances available as resource.

        :return: the tuple with the instance names

        >>> len(list(Instance.list_resources()))
        553
        """
        return _INSTANCES

    @staticmethod
    def from_resource(name: str) -> "Instance":
        """
        Load an instance from a resource.

        :param name: the name string
        :return: the instance

        >>> Instance.from_resource("a01").to_compact_str()
        'a01;2;2750;1220;463,386,18;1680,420,6'
        >>> Instance.from_resource("a07").to_compact_str()
        'a07;5;2750;1220;706,286,8;706,516,16;986,433,10;1120,486,\
12;1120,986,12'
        >>> Instance.from_resource("a08") is Instance.from_resource("a08")
        True
        >>> Instance.from_resource("a08") is Instance.from_resource("a09")
        False
        """
        if not isinstance(name, str):
            raise type_error(name, "name", str)
        container: Final = Instance.from_resource
        inst_attr: Final[str] = f"__inst_{name}"
        if hasattr(container, inst_attr):  # instance loaded?
            return cast(Instance, getattr(container, inst_attr))
        text_attr: Final[str] = f"__text_{INSTANCES_RESOURCE}"
        text: list[str]
        total_attr: Final[str] = "__total_insts"
        if hasattr(container, text_attr):  # ok, we got the text already
            text = cast(list[str], getattr(container, text_attr))
        else:  # the first time we load the text
            with resources.open_text(package=str(__package__),
                                     resource=INSTANCES_RESOURCE) as stream:
                text = [line for line in (line_raw.strip() for line_raw
                                          in stream.readlines())
                        if len(line) > 0]
            setattr(container, text_attr, text)
            setattr(container, total_attr, 0)  # so far, no instance

        imax: int = len(text)
        imin: int = 0
        while imin <= imax:  # binary search for instance
            imid: int = (imax + imin) // 2
            line: str = text[imid]
            idx: int = line.index(CSV_SEPARATOR)
            prefix: str = line[0:idx]
            if prefix == name:
                instance = Instance.from_compact_str(line)
                setattr(container, inst_attr, instance)
                got: int = getattr(container, total_attr)
                got = got + 1
                if got >= len(_INSTANCES):  # got all instances, can del text
                    delattr(container, total_attr)
                    delattr(container, text_attr)
                return instance
            if prefix < name:
                imin = imid + 1
            else:
                imax = imid - 1
        raise ValueError(f"instance {name!r} not found.")



"""
Applications of Metaheuristic Optimization in Python.

Currently, the following applications are implemented:

- :mod:`~moptipyapps.binpacking2d` provides methods to solve two-dimensional
  bin packing instances.

The following additional tools are implemented:

- :mod:`~moptipyapps.tests` offers unit tests to try out optimization
  algorithms and other instances of :mod:`~moptipy.api.component` on the
  different problems that are provided above.
- :mod:`~moptipyapps.shared` offers shared constants and tools.
"""
from moptipy.utils.sys_info import add_dependency

add_dependency("moptipyapps")



"""
A (1+1) FEA for the TSP using the reversal move.

A (1+1) FEA is the same as the (1+1) EA with Frequency Fitness Assignment.
The (1+1) EA using the same search operator as this algorithm here is
implemented in module :mod:`~moptipyapps.tsp.ea1p1_revn`.
The algorithm implemented here is the same as the basic (1+1) FEA with `rev`
operator in the paper [1].

The original version of this code has been contributed by Mr. Tianyu LIANG
(梁天宇), <liangty@stu.hfuu.edu.cn> a Master's student at the Institute of
Applied Optimization (应用优化研究所, http://iao.hfuu.edu.cn) of the School
of Artificial Intelligence and Big Data (人工智能与大数据学院) at Hefei
University (合肥学院) in Hefei, Anhui, China (中国安徽省合肥市) under the
supervision of Prof. Dr. Thomas Weise (汤卫思教授).

1. Tianyu Liang, Zhize Wu, Jörg Lässig, Daan van den Berg, and Thomas Weise.
   Solving the Traveling Salesperson Problem using Frequency Fitness
   Assignment. In Hisao Ishibuchi, Chee-Keong Kwoh, Ah-Hwee Tan, Dipti
   Srinivasan, Chunyan Miao, Anupam Trivedi, and Keeley A. Crockett, editors,
   *Proceedings of the IEEE Symposium on Foundations of Computational
   Intelligence (IEEE FOCI'22)*, part of the *IEEE Symposium Series on
   Computational Intelligence (SSCI 2022)*. December 4-7, 2022, Singapore,
   pages 360-367. IEEE. https://doi.org/10.1109/SSCI51031.2022.10022296.
2. Pedro Larrañaga, Cindy M. H. Kuijpers, Roberto H. Murga, Iñaki Inza, and
   S. Dizdarevic. Genetic Algorithms for the Travelling Salesman Problem: A
   Review of Representations and Operators. *Artificial Intelligence Review,*
   13(2):129-170, April 1999. Kluwer Academic Publishers, The Netherlands.
   https://doi.org/10.1023/A:1006529012972.
"""

from typing import Callable, Final, cast

import numba  # type: ignore
import numpy as np
from moptipy.api.algorithm import Algorithm
from moptipy.api.process import Process
from moptipy.utils.logger import CSV_SEPARATOR, KeyValueLogSection
from moptipy.utils.nputils import DEFAULT_INT, array_to_str
from moptipy.utils.types import type_error
from numpy.random import Generator

from moptipyapps.shared import SCOPE_INSTANCE
from moptipyapps.tsp.instance import Instance


@numba.njit(nogil=True, cache=True, inline="always")
def rev_if_h_not_worse(i: int, j: int, n_cities: int, dist: np.ndarray,
                       h: np.ndarray, x: np.ndarray, y: int) -> int:
    """
    Apply a reversal move if its tour length frequency is not worse.

    :param i: the first, smaller index
    :param j: the second, larger index
    :param n_cities: the number of cities
    :param dist: the problem instance
    :param h: the frequency table
    :param x: the candidate solution
    :param y: the tour length
    """
    xi: Final[int] = x[i]  # the value of x at index i
    xim1: Final[int] = x[((i - 1) + n_cities) % n_cities]  # x[i - 1]
    xj: Final[int] = x[j]  # the value of x at index j
    xjp1: Final[int] = x[(j + 1) % n_cities]  # x[j + 1], but index wrapped

    # compute the difference in tour length if we would apply the move
    dy: Final[int] = (dist[xim1, xj] + dist[xi, xjp1]
                      - dist[xim1, xi] - dist[xj, xjp1])
    y2: Final[int] = int(y + dy)  # and compute the new tour length
    h[y] += 1  # update frequency of the tour length of the current solution
    h[y2] += 1  # update frequency of the tour length of the new solution
    if h[y2] <= h[y]:  # this move does not make the frequency worse?
        # so we reverse the sequence from i to j in the solution
        if i == 0:  # deal with the special case that i==0
            x[0:j + 1:1] = x[j::-1]
        else:  # the normal case that i > 0
            x[i:j + 1:1] = x[j:i - 1:-1]
        return y2  # return new tour length
    return y  # return old tour length


def log_frequency_table(h: np.ndarray, process: Process) -> None:
    """
    Log the frequency table to a given process.

    The frequency table is logged as a string of the form
    `a;b;h[a];h[a+1];...;h[b-1];h[b]` into a section with name `H`.
    `a` is the first objective value that was encountered, `b` the last one.
    Normally, `a` should be the tour length of the best tour that was
    discovered.

    :param h: the frequency table
    :param process: the process
    """
    i: int = 0
    n_h: Final[int] = len(h)
    for i in range(cast(int, process.get_best_f()), n_h):
        if h[i] > 0:  # find first non-zero element
            break
    j: int = n_h
    for j in range(n_h - 1, i, -1):
        if h[j] > 0:  # find last non-zero element
            break

    s: str  # the string with the data to log
    if i <= 2:  # can we use the dirty trick below?
        h[i - 2] = i  # dirty trick to easily store indices of first and
        h[i - 1] = j  # last non-zero items
        s = array_to_str(h[i - 2:j + 1])
    else:  # no ... ok, then we just write it as-is
        s = f"{i}{CSV_SEPARATOR}{j}{CSV_SEPARATOR}{array_to_str(h[i:j + 1])}"

    process.add_log_section("H", s)


class TSPFEA1p1revn(Algorithm):
    """A (1+1) FEA using the reversal operator for the TSP."""

    def __init__(self, instance: Instance) -> None:
        """
        Initialize the RLS algorithm for the TSP with reversal move.

        :param instance: the problem instance
        """
        super().__init__()
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        self.instance: Final[Instance] = instance

    def solve(self, process: Process) -> None:
        """
        Apply a (1+1) FEA optimization process with reversing operator.

        :param process: the process instance which provides random numbers,
            functions for creating, copying, and evaluating solutions, as well
            as the termination criterion
        """
        random: Final[Generator] = process.get_random()
        # set up the fast calls
        register: Final[Callable[[np.ndarray, int], int]] =\
            cast(Callable[[np.ndarray, int], int], process.register)
        should_terminate: Final[Callable[[], bool]] = process.should_terminate
        ri: Final[Callable[[int], int]] = random.integers

        instance: Final[Instance] = self.instance  # get the instance
        h: Final[np.ndarray] = np.zeros(  # allocate the frequency table
            instance.tour_length_upper_bound + 1, DEFAULT_INT)
        n: Final[int] = instance.n_cities  # get the number of cities
        x: Final[np.ndarray] = process.create()  # create the solution
        x[:] = range(n)  # fill array with 0..n
        random.shuffle(x)  # randomly generate an initial solution

        y: int = cast(int, process.evaluate(x))  # get length of first tour
        nm1: Final[int] = n - 1  # need n-1 in the loop for the random numbers
        nm2: Final[int] = n - 12  # we need this to check the move indices
        while not should_terminate():
            i = ri(nm1)  # get the first index
            j = ri(nm1)  # get the second index
            if i > j:  # ensure that i <= j
                i, j = j, i  # swap indices if i > j
            if (i == j) or ((i == 0) and (j == nm2)):
                continue  # either a nop or a complete reversal
            y = rev_if_h_not_worse(i, j, n, instance, h, x, y)  # move
            register(x, y)  # register the objective value

        # we log the frequency table at the very end of the run
        log_frequency_table(h, process)

    def __str__(self):
        """
        Get the name of this algorithm.

        This name is then used in the directory path and file name of the
        log files.

        :returns: "tsp_fea1p1_revn"
        :retval "tsp_fea1p1_revn": always
        """
        return "tsp_fea1p1_revn"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the algorithm to the given logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        with logger.scope(SCOPE_INSTANCE) as kv:
            self.instance.log_parameters_to(kv)



"""
An improved bottom left encoding by Liu and Teng extended to multiple bins.

Here we provide an implementation of the improved bottom left encoding by Liu
and Teng [1], but extended to bins with limited height. If the height of the
bin is a limiting factor, then our implementation will automatically use
multiple bins. Another variant is given in
:mod:`moptipyapps.binpacking2d.ibl_encoding_1`.

An instance :mod:`~moptipyapps.binpacking2d.instance` of the
two-dimensional bin packing problem defines a set of objects to be packed
and a bin size (width and height). Each object to be packed has itself a
width and a height as well as a repetition counter, which is `1` if the object
only occurs a single time and larger otherwise (i.e., if the repetition
counter is `5`, the object needs to be packaged five times).

The encoding receives signed permutations with repetitions as input. Each
element of the permutation identifies one object from the bin packing
instance. Each such object ID must occur exactly as often as the repetition
counter of the object in the instance data suggest. But the ID may also occur
negated, in which case the object is supposed to rotated by 90°.

Now our encoding processes such a permutation from beginning to end. It starts
with an empty bin `1`. Each object is first placed with its right end at the
right end of the first bin and with its bottom line exactly at the top of the
bin, i.e., outside of the bin. Then, in each step, we move the object as far
down as possible. Then, we move it to the left as far as possible, but we
immediately stop if there was another chance to move the object down. In
other words, downward movements are preferred over left movements. This is
repeated until no movement of the object is possible anymore.

Once the object cannot be moved anymore, we check if it is fully inside the
bin. If yes, then the object is included in the bin and we continue with the
next object. If not, it does not fit into the bin.

This is the "Improved Bottom Left" heuristic by Liu and Teng [1].

If the object does not fit into the first bin and we already "opened" a second
bin, then we try to place it into the second bin using the same procedure. And
then into the third bin if that does not work out, and so on. Until we have
tried unsuccessfully all the bins that we have opened.

In this case, we "open" the next bin and we place the object at the
bottom-left corner of a new bin. Then we continue with the next object, again
trying to put it into the first bin, then the second bin, and so on.

This is different from the first variant of this encoding implemented in file
:mod:`moptipyapps.binpacking2d.ibl_encoding_2`, which always and only
tries to put objects into the last bin that was opened (and moves to a new bin
if that does not work out). That variant of the encoding is therefore faster
than the one here, but the one here tends to yield better packings.

The problem with this is that we need to basically first try to put the object
into the first bin. For this we need to look at *all* the objects that we have
already placed, because it could be that we already have one object in the
first bin, then one in the second bin that did not fit into the first bin,
then one smaller object in the first bin again, and so on. If our new object
does not fit into the first bin, then we need to do the same with the second
bin, and so on. So for every bin we try, we need to look at all objects already
placed. And the number of bins we could try could be equal to the number of
objects that we have already placed (if each object occupies one bin alone).
So we have a worst case complexity of O(n ** 2) for placing one object. And we
do this for all objects, so we would have a O(n ** 3) overall complexity.
Well, actually, it is worse: Because we repeat the process of looking at all
the objects several times while moving our new item to the left and down and
to the left and down. So I suspect that we actually have O(n ** 4).
That is annoying.

We try to alleviate this a little bit by remembering, for each bin, the index
of the first object that we put in there and the index of the last object we
put in there. Now within these two indices, there also might be objects that
we placed into other bins. But for a very little overhead (remembering two
values per bin), we have a certain chance to speed up the process in several
situations. For instance, the worst case from above, that each object occupies
exactly one bin by itself becomes easier because we would only look at one
already placed object per bin.

This procedure has originally been developed and implemented by Mr. Rui ZHAO
(赵睿), <zr1329142665@163.com>, a Master's student at the Institute of Applied
Optimization (应用优化研究所, http://iao.hfuu.edu.cn) of the School of
Artificial Intelligence and Big Data (人工智能与大数据学院) at Hefei University
(合肥学院) in Hefei, Anhui, China (中国安徽省合肥市) under the supervision of
Prof. Dr. Thomas Weise (汤卫思教授).

1. Dequan Liu and Hongfei Teng. An Improved BL-Algorithm for Genetic Algorithm
   of the Orthogonal Packing of Rectangles. European Journal of Operational
   Research. 112(2):413-420. January (1999).
   https://doi.org/10.1016/S0377-2217(97)00437-2.
   http://www.paper.edu.cn/scholar/showpdf/MUT2AN0IOTD0Mxxh.
"""
from typing import Final

import numba  # type: ignore
import numpy as np
from moptipy.api.encoding import Encoding
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.types import type_error

from moptipyapps.binpacking2d.instance import (
    IDX_HEIGHT,
    IDX_WIDTH,
    Instance,
)
from moptipyapps.binpacking2d.packing import (
    IDX_BIN,
    IDX_BOTTOM_Y,
    IDX_ID,
    IDX_LEFT_X,
    IDX_RIGHT_X,
    IDX_TOP_Y,
    Packing,
)
from moptipyapps.shared import SCOPE_INSTANCE


@numba.njit(nogil=True, cache=True, inline="always")
def __move_down(packing: np.ndarray, bin_id: int,
                bin_start: int, bin_end: int, i1: int) -> bool:
    """
    Move the box at index `i1` down as far as possible in the current bin.

    `bin_start` is the index of the first object that has already been placed
    in the current bin `bin_id`. It always holds that `i1 >= bin_start`. In
    the case that `i1 == bin_start`, then we can move the object directly to
    the bottom of the bin without any issue.
    If `i1 > bin_start` we iterate over all objects at indices
    `bin_start...bin_end - 1` whose bin equals `bin`. We first set `min_down`
    to the bottom-y coordinate of the box, because this is how far down we can
    move at most.

    Then, for each of the objects already placed in the bin `bin_id`, we check
    if there is any intersection of the horizontal with the current box. If
    there is no intersection *or* if the object is not already above the
    current box, then the object will not influence the downward movement of
    our object. If there is an intersection, then we cannot move the current
    box deeper than the top-y coordinate of the other box.

    *Only* the box at index `i1` is modified and if it is modified, this
    function will return `True`.

    :param packing: the packing under construction
    :param bin_id: the bin ID
    :param bin_start: the starting index of the current bin
    :param bin_end: the exclusive end index of the current bin
    :param i1: the index of the current box
    :return: `True` if the object was moved down, `False` if the object cannot
        be moved down any further because either it has reached the bottom or
        because it would intersect with other objects

    >>> # itemID, binID, left-x, bottom-y, right-x, top-y
    >>> r = np.array([[1, 1, 10, 20, 30, 40],
    ...               [2, 1, 30, 30, 50, 60],
    ...               [3, 1, 40, 100, 60, 200]])
    >>> __move_down(r, 1, 0, 2, 2)  # try to move down the box at index 2
    True
    >>> print(r[2, :])
    [  3   1  40  60  60 160]
    >>> __move_down(r, 1, 0, 2, 2)  # try to move down the box at index 2
    False
    >>> __move_down(r, 1, 0, 1, 1)  # try to move down the box at index 1
    True
    >>> print(r[1, :])
    [ 2  1 30  0 50 30]
    >>> __move_down(r, 1, 0, 1, 1)  # try to move down the box at index 1
    False
    >>> __move_down(r, 1, 0, 2, 2)  # try to move down the box at index 2
    True
    >>> print(r[2, :])
    [  3   1  40  30  60 130]
    >>> __move_down(r, 1, 0, 2, 2)  # try to move down the box at index 2
    False
    >>> __move_down(r, 1, 0, 0, 0)
    True
    >>> print(r[0, :])
    [ 1  1 10  0 30 20]
    >>> __move_down(r, 1, 0, 0, 0)
    False
    """
    # load the coordinates of i1 into local variables to speed up computation
    packing_i1_left_x: Final[int] = packing[i1, IDX_LEFT_X]
    packing_i1_bottom_y: Final[int] = packing[i1, IDX_BOTTOM_Y]
    packing_i1_right_x: Final[int] = packing[i1, IDX_RIGHT_X]
    packing_i1_top_y: Final[int] = packing[i1, IDX_TOP_Y]
    min_down: int = packing_i1_bottom_y  # maximum move: down to bottom
    for i0 in range(bin_start, bin_end):  # iterate over items in current bin
        # An intersection exists if the right-x of an existing box is larger
        # than the left-x of the new box AND if the left-x of the existing box
        # is less than the right-x of the new box.
        # Only intersections matter and *only* with objects in the same bin
        # and only if the existing box is not already above the new box.
        if (packing[i0, IDX_BIN] == bin_id) and \
                (packing[i0, IDX_RIGHT_X] > packing_i1_left_x) and \
                (packing[i0, IDX_LEFT_X] < packing_i1_right_x) and \
                (packing[i0, IDX_BOTTOM_Y] < packing_i1_top_y):
            # The object would horizontally intersect with the current object
            diff: int = packing_i1_bottom_y - packing[i0, IDX_TOP_Y]
            if diff < min_down:  # we can move down only a shorter distance
                min_down = diff  # ok, a shorter distance was found
    if min_down > 0:  # Can we move down? If yes, update box.
        packing[i1, IDX_BOTTOM_Y] = packing_i1_bottom_y - min_down
        packing[i1, IDX_TOP_Y] = packing_i1_top_y - min_down
        return True
    return False


@numba.njit(nogil=True, cache=True, inline="always")
def __move_left(packing: np.ndarray, bin_id: int,
                bin_start: int, bin_end: int, i1: int) -> bool:
    """
    Move the box at index `i1` left as far as possible in the current bin.

    This function moves a box to the left without changing its vertical
    position. It is slightly more tricky than the downwards moving function,
    because in the improved bottom left heuristic, downward moves are
    preferred compared to left moves. This means that the box needs to be
    stopped when reaching the edge of a box on whose top it sits. Of course,
    we only consider other objects inside the current bin
    (with ID `bin_start`).

    This function is to be called *after* `__move_down` and in an alternating
    fashion.

    *Only* the box at index `i1` is modified and if it is modified, this
    function will return `True`.

    :param packing: the packing under construction
    :param bin_id: the bin ID
    :param bin_start: the starting index of the current bin
    :param bin_end: the exclusive end index of the current bin
    :param i1: the index of the current box
    :return: `True` if the object was moved down, `False` if the object cannot
        be moved down any further because either it has reached the bottom or
        because it would intersect with other objects

    >>> # itemID, binID, left-x, bottom-y, right-x, top-y
    >>> r = np.array([[1, 1,  0,  0, 30, 10],
    ...               [2, 1, 35,  0, 45, 30],
    ...               [3, 1,  0, 10, 10, 20],
    ...               [4, 1, 40, 30, 50, 40]])
    >>> __move_left(r, 1, 0, 3, 3)
    True
    >>> print(r[3, :])
    [ 4  1 25 30 35 40]
    >>> __move_left(r, 1, 0, 3, 3)
    True
    >>> print(r[3, :])
    [ 4  1  0 30 10 40]
    >>> r[3, :] = [4, 1, 25, 10, 35, 20]
    >>> __move_left(r, 1, 0, 3, 3)
    True
    >>> print(r[3, :])
    [ 4  1 10 10 20 20]
    >>> __move_left(r, 1, 0, 3, 3)
    False
    >>> # itemID, binID, left-x, bottom-y, right-x, top-y
    >>> r = np.array([[1, 1,  0,  0, 10, 2],
    ...               [2, 1, 10,  0, 20, 5],
    ...               [3, 1,  8,  2, 10, 4]])
    >>> __move_left(r, 0, 0, 3, 2)
    True
    >>> print(r[2, :])
    [3 1 0 2 2 4]
    """
    packing_i1_left_x: Final[int] = packing[i1, IDX_LEFT_X]
    packing_i1_bottom_y: Final[int] = packing[i1, IDX_BOTTOM_Y]
    packing_i1_right_x: Final[int] = packing[i1, IDX_RIGHT_X]
    packing_i1_top_y: Final[int] = packing[i1, IDX_TOP_Y]
    min_left: int = packing_i1_left_x
    for i0 in range(bin_start, bin_end):
        # consider only objects that are not yet behind us and are in
        # the same bin may intersect
        if (packing[i0, IDX_BIN] != bin_id) \
                or (packing[i0, IDX_LEFT_X] >= packing_i1_right_x):
            continue  # the object is already behind us, so it can be ignored
        if (packing[i0, IDX_RIGHT_X] > packing_i1_left_x) \
                and (packing[i0, IDX_LEFT_X] < packing_i1_right_x):
            # we have a horizontal intersection with a box below
            if packing[i0, IDX_TOP_Y] == packing_i1_bottom_y:
                # only consider those the box *directly* below and move the
                # right end of the new box to the left end of that box below
                diff = packing_i1_right_x - packing[i0, IDX_LEFT_X]
                if diff < min_left:
                    min_left = diff
        elif (packing_i1_top_y > packing[i0, IDX_BOTTOM_Y]) \
                and (packing_i1_bottom_y < packing[i0, IDX_TOP_Y]):
            diff = packing_i1_left_x - packing[i0, IDX_RIGHT_X]
            if diff < min_left:
                min_left = diff
    if min_left > 0:
        # move the box to the left
        packing[i1, IDX_LEFT_X] = packing_i1_left_x - min_left
        packing[i1, IDX_RIGHT_X] = packing_i1_right_x - min_left
        return True
    return False


@numba.njit(nogil=True, cache=True, inline="always")
def _decode(x: np.ndarray, y: np.ndarray, instance: np.ndarray,
            bin_width: int, bin_height: int, bin_starts: np.ndarray,
            bin_ends: np.ndarray) -> int:
    """
    Decode a (signed) permutation to a packing.

    The permutation is processed from the beginning to the end.
    Each element identifies one object by its ID. If the ID is negative,
    the object will be inserted rotated by 90°. If the ID is positive, the
    object will be inserted as is.

    The absolute value of the ID-1 will be used to look up the width and
    height of the object in the `instance` data. If the object needs to be
    rotated, width and height will be swapped.

    Each object is, at the beginning, placed with its right side at the right
    end of the *first* bin. The bottom line of the object is initially put on
    top of the bin, i.e., initially the object is outside of the bin.

    Then, the object is iteratively moved downward *in the bin* as far as
    possible. Once it reaches another object, we move it to the left until
    either its right side reaches the left end of the object beneath it or
    until its left side touches another object. Then we try to move the object
    down again, and so on.

    Once the object can no longer be moved down, we check if it is now fully
    inside of *the bin*. If yes, then good, the object's bin index is set to
    the ID of the current bin. If not, then we cannot place the object into
    this bin. In this case, we move to the second bin an repeat the process.
    If the object fits into the second bin, then it is placed in there and
    we are finished. If not, we try the third bin, and so on.

    If we cannot fit the object into any of the bins that already have other
    objects, then we allocate a new empty bin. We move the object to the
    bottom-left corner of this bin. In other words, the left side of the
    object touches the left side of the bin, i.e., is `0`. The bottom-line
    of the object is also the bottom of the bin, i.e., has coordinate `0` as
    well.

    :param x: a possibly signed permutation
    :param y: the packing object
    :param instance: the packing instance data
    :param bin_width: the bin width
    :param bin_height: the bin height
    :param bin_starts: a temporary index holder array for bin starts
    :param bin_ends: a temporary index holder array for bin ends
    :returns: the number of bins

    As example, we use a slightly modified version (we add more objects so we
    get to see the use of a second bin) of Figure 2 of the Liu and Teng paper
    "An Improved BL-Algorithm for Genetic Algorithm of the Orthogonal Packing
    of Rectangles."

    >>> # [width, height, repetitions]
    >>> inst = np.array([[10, 20, 5], [5, 5, 5]])
    >>> # [id = plain, -id = rotated]
    >>> xx = np.array([1, -1, 2, -2, 1, -2, -2, -1, -1, 2])
    >>> # [id, bin, left-x, bottom-y, right-x, top-y] ...
    >>> yy = np.empty((10, 6), int)
    >>> binstarts = np.empty(10, int)
    >>> binends = np.empty(10, int)
    >>> print(_decode(xx, yy, inst, 30, 30, binstarts, binends))
    2
    >>> print(yy[0, :])
    [ 1  1  0  0 10 20]
    >>> print(yy[1, :])
    [ 1  1 10  0 30 10]
    >>> print(yy[2, :])
    [ 2  1 10 10 15 15]
    >>> print(yy[3, :])
    [ 2  1 15 10 20 15]
    >>> print(yy[4, :])
    [ 1  1 20 10 30 30]
    >>> print(yy[5, :])
    [ 2  1 10 15 15 20]
    >>> print(yy[6, :])
    [ 2  1 15 15 20 20]
    >>> print(yy[7, :])
    [ 1  1  0 20 20 30]
    >>> print(yy[8, :])
    [ 1  2  0  0 20 10]
    >>> print(yy[9, :])
    [ 2  2 20  0 25  5]
    """
    w: int  # the width of the current object
    h: int  # the height of the current object
    use_id: int  # the id of the current object
    bin_id: int = 1  # the id of the current bin
    bin_starts[0] = 0
    bin_ends[0] = 0
    for i, item_id in enumerate(x):  # iterate over all objects
        if item_id < 0:  # object should be rotated
            use_id = -(item_id + 1)  # get absolute id - 1
            w = instance[use_id, IDX_HEIGHT]  # width = height (rotated!)
            h = instance[use_id, IDX_WIDTH]   # height = width (rotated!)
        else:  # the object will not be rotated
            use_id = item_id - 1   # id - 1
            w = instance[use_id, IDX_WIDTH]  # get width
            h = instance[use_id, IDX_HEIGHT]  # get height

# It could be that an object is too wide or too high for the bin in its
# current rotation even if the bin was empty entirely. In this case, we simply
# force-rotate it. A bin packing instance will not permit objects that do not
# fit into the bin in any rotation. So if the object does not fit in its
# current rotation, it must fit if we simply rotate it by 90°.
        if (w > bin_width) or (h > bin_height):
            w, h = h, w

# At first, the object's right corner is at the right corner of the bin.
# The object sits exactly at the top of the bin, i.e., its bottom line
# is the top line of the bin.
        y[i, IDX_ID] = use_id + 1  # the id of the object

        not_found: bool = True
        for item_bin in range(1, bin_id + 1):  # iterate over all bins in use
            bin_start = bin_starts[item_bin - 1]  # index of first item in bin
            bin_end = bin_ends[item_bin - 1]  # index after last item in bin
            y[i, IDX_LEFT_X] = bin_width - w  # the left end
            y[i, IDX_BOTTOM_Y] = bin_height  # object sits on top of bin
            y[i, IDX_RIGHT_X] = bin_width  # object ends at right end of bin
            y[i, IDX_TOP_Y] = bin_height + h  # top of object is outside of bin

            while __move_down(y, item_bin, bin_start, bin_end, i) \
                    or __move_left(y, item_bin, bin_start, bin_end, i):
                pass  # loop until object can no longer be moved

# Check if the object is completely inside the bin. If yes, we can put it
# there and stop searching.
            if (y[i, IDX_RIGHT_X] <= bin_width) \
                    and (y[i, IDX_TOP_Y] <= bin_height):
                not_found = False  # we found a placement, outer loop can stop
                y[i, IDX_BIN] = item_bin  # set the items bin
                bin_ends[item_bin - 1] = i + 1  # index after last item in bin
                break

# If the object is not fully inside the current bin, we move to a new bin.
        if not_found:  # we did not find a spot in any of the bins
            bin_starts[bin_id] = i  # set the starting index of the new bin
            bin_ends[bin_id] = i + 1  # set the end index of the new bin
            bin_id = bin_id + 1  # step to the next bin
            y[i, IDX_LEFT_X] = 0  # the object goes to the left end of the bin
            y[i, IDX_BOTTOM_Y] = 0  # the object goes to the bottom of the bin
            y[i, IDX_RIGHT_X] = w  # so its right end is its width
            y[i, IDX_TOP_Y] = h  # and its top end is its height
            y[i, IDX_BIN] = bin_id  # store the bin id
    return int(bin_id)  # return the total number of bins


class ImprovedBottomLeftEncoding2(Encoding):
    """An Improved Bottem Left Encoding by Liu and Teng for multiple bins."""

    def __init__(self, instance: Instance) -> None:
        """
        Instantiate the improved best first encoding.

        :param instance: the packing instance
        """
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        #: the internal instance reference
        self.__instance: Final[Instance] = instance
        #: the temporary array for holding bin start indexes
        self.__bin_starts: Final[np.ndarray] = np.empty(
            instance.n_items, instance.dtype)
        #: the temporary array for holding bin end indexes
        self.__bin_ends: Final[np.ndarray] = np.empty(
            instance.n_items, instance.dtype)

    def decode(self, x: np.ndarray, y: Packing) -> None:
        """
        Map a potentially signed permutation to a packing.

        :param x: the array
        :param y: the Gantt chart
        """
        y.n_bins = _decode(x, y, self.__instance, self.__instance.bin_width,
                           self.__instance.bin_height, self.__bin_starts,
                           self.__bin_ends)

    def __str__(self) -> str:
        """
        Get the name of this encoding.

        :return: `"ibf2"`
        :rtype: str
        """
        return "ibf2"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        with logger.scope(SCOPE_INSTANCE) as kv:
            self.__instance.log_parameters_to(kv)
